using UnityEngine;
using System;
using System.Collections.Generic;
using System.Text;
using System.Net;
using System.Net.Sockets;

namespace VeeRifter
{
	public class Kinect360ToVAM : MVRScript
    {
        private const string pluginName = "Kinect360ToVAM";
		private const string pluginAuthor = "VeeRifter";
		private const string pluginVersion = "v1.2";
		private const string pluginDate = "15-11-21";
		private const string aboutText = "\n\n<b><size=40><color=brown>" + pluginName + " " + pluginVersion + "</color></size></b>\n<i>" + pluginAuthor + "\n" + pluginDate + "</i>";
		private int listenPort = 6360;
		private IPEndPoint ipep;
		private IPEndPoint sender;
		private EndPoint remote;
		private Socket udpSocket;
		private byte[] inBuffer = new byte[640];
		private byte[] connectPacket = {69};
		private byte[] disconnectPacket = {96};

		private Body kinect;
		private FreeControllerV3[] jointControls;
		private UIDynamicButton selectAllButton;
		private UIDynamicButton selectNoneButton;
		private UIDynamicButton calibrateButton;
		private UIDynamicButton connectButton;
		private UIDynamicButton disconnectButton;
		private UIDynamicButton resetButton;
		private UIDynamicTextField serverStatus;
		private JSONStorableString status;
		private JSONStorableBool lToe;
		private JSONStorableBool lFoot;
		private JSONStorableBool lKnee;
		private JSONStorableBool lThigh;
		private JSONStorableBool lHand;
		private JSONStorableBool lElbow;
		private JSONStorableBool lShoulder;
		private JSONStorableBool rToe;
		private JSONStorableBool rFoot;
		private JSONStorableBool rKnee;
		private JSONStorableBool rThigh;
		private JSONStorableBool rHand;
		private JSONStorableBool rElbow;
		private JSONStorableBool rShoulder;
		private JSONStorableBool hip;
		private JSONStorableBool chest;
		private JSONStorableBool pelvis;
		private JSONStorableBool head;
		private JSONStorableBool tracking;
		private bool connected = false;
		private bool calibrated = false;
		private bool selectNone = false;
		private bool timerEnabled = false;
		private float calibrateTimer = 0.0f;
		private float watchdogTimer = 0.0f;
		
		public override void Init()
		{
			try
			{
				if (containingAtom.type != "Person")
				{
					SuperController.LogError($"This plugin can only be used with a 'Person' atom.");
					return;
				}
								
				udpSocket = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
				ipep = new IPEndPoint(IPAddress.Loopback, listenPort);
				sender = new IPEndPoint(IPAddress.Any, 0);				
				remote = (EndPoint)sender;

				// UI left side
				JSONStorableString about = new JSONStorableString("About", aboutText);
				UIDynamicTextField dtfAbout = CreateTextField(about, false);
				dtfAbout.UItext.alignment = TextAnchor.MiddleCenter;
				dtfAbout.UItext.supportRichText = true;
				dtfAbout.UItext.fontSize = 30;		
				dtfAbout.UItext.color = Color.black;
				dtfAbout.height = 230;

				tracking = new JSONStorableBool("Kinect Tracking On", true);
				tracking.storeType = JSONStorableParam.StoreType.Full;
				RegisterBool(tracking);				
				CreateToggle(tracking, false);

				selectAllButton = CreateButton("Select All", false);
				selectAllButton.buttonColor = Color.green;
				selectAllButton.button.onClick.AddListener(SelectAll);
				
				head = CreateBoolCheckbox("Head", true, false);
				chest = CreateBoolCheckbox("Chest", true, false);
				lShoulder = CreateBoolCheckbox("Left Shoulder", true, false);
				lElbow = CreateBoolCheckbox("Left Elbow", true, false);
				lHand = CreateBoolCheckbox("Left Hand", true, false);
				lThigh = CreateBoolCheckbox("Left Thigh", true, false);
				lKnee = CreateBoolCheckbox("Left Knee", true, false);
  				lFoot = CreateBoolCheckbox("Left Foot", true, false);
				lToe = CreateBoolCheckbox("Left Toe", true, false);

				UIDynamic leftSpacer1 = CreateSpacer(false);
				leftSpacer1.height = 50f;

				calibrateButton = CreateButton("Calibrate", false);
				calibrateButton.buttonColor = Color.blue;
				calibrateButton.textColor = Color.white;
				calibrateButton.button.onClick.AddListener(startCountdown);

				// UI right side	
				connectButton = CreateButton("Connect", true);
				connectButton.buttonColor = Color.green;
				connectButton.button.onClick.AddListener(Connect);

				disconnectButton = CreateButton("Disconnect", true);
				disconnectButton.buttonColor = Color.red;
				disconnectButton.textColor = Color.white;
				disconnectButton.button.onClick.AddListener(Disconnect);

				status = new JSONStorableString("Status", "\nServer Status: <color=red>Disconnected</color>");
				UIDynamicTextField serverStatus = CreateTextField(status, true);
				serverStatus.UItext.alignment = TextAnchor.LowerCenter;
				serverStatus.UItext.fontStyle = FontStyle.Italic;
				serverStatus.UItext.supportRichText = true;
				serverStatus.height = 50;

				UIDynamic rightSpacer1 = CreateSpacer(true);
				rightSpacer1.height = 50f;

				selectNoneButton = CreateButton("Select None", true);
				selectNoneButton.buttonColor = Color.red;
				selectNoneButton.textColor = Color.white;				
				selectNoneButton.button.onClick.AddListener(SelectNone);

				pelvis = CreateBoolCheckbox("Pelvis", true, true);
				hip = CreateBoolCheckbox("Hip", true, true);	
				rShoulder = CreateBoolCheckbox("Right Shoulder", true, true);				
				rElbow = CreateBoolCheckbox("Right Elbow", true, true);				
				rHand = CreateBoolCheckbox("Right Hand", true, true);				
				rThigh = CreateBoolCheckbox("Right Thigh", true, true);				
				rKnee = CreateBoolCheckbox("Right Knee", true, true);				
  				rFoot = CreateBoolCheckbox("Right Foot", true, true);              
				rToe = CreateBoolCheckbox("Right Toe", true, true);				

				UIDynamic rightSpacer2 = CreateSpacer(true);
				rightSpacer2.height = 50f;

				resetButton = CreateButton("Reset", true);
				resetButton.buttonColor = Color.yellow;
				resetButton.textColor = Color.black;		
				resetButton.button.onClick.AddListener(calReset);

				kinect = new Body();
				jointControls = new FreeControllerV3[20];

				var controllers = containingAtom.GetComponentsInChildren<FreeControllerV3>(true);
				foreach (var control in controllers)
				{
					if (control == null || control.followWhenOff == null || control.control == null)
						continue;
					
					switch (control.name)
					{
						case "hipControl":
							jointControls[0] = control;
							break;
						case "pelvisControl":
							jointControls[1] = control;
							break;
						case "chestControl":
							jointControls[2] = control;
							break;
						case "headControl":
							jointControls[3] = control;
							break;
						case "lArmControl":
							jointControls[4] = control;
							break;
						case "lElbowControl":
							jointControls[5] = control;
							break;
						case "lHandControl":
							jointControls[6] = control;
							break;
						case "lhand":
							jointControls[7] = control;
							break;
						case "rArmControl":
							jointControls[8] = control;
							break;
						case "rElbowControl":
							jointControls[9] = control;
							break;
						case "rHandControl":
							jointControls[10] = control;
							break;
						case "rhand":
							jointControls[11] = control;
							break;
						case "lThighControl":
							jointControls[12] = control;
							break;
						case "lKneeControl":
							jointControls[13] = control;
							break;
						case "lFootControl":
							jointControls[14] = control;
							break;
						case "lToeControl":
							jointControls[15] = control;
							break;
						case "rThighControl":
							jointControls[16] = control;
							break;
						case "rKneeControl":
							jointControls[17] = control;
							break;
						case "rFootControl":
							jointControls[18] = control;
							break;
						case "rToeControl":
							jointControls[19] = control;
							break;
					}
				}
				setTrackedJoints(true);
			}
			catch (System.Exception e)
			{
				SuperController.LogError("Exception caught in Init(): " + e);
			}
		}

		void Update()
		{
			if (timerEnabled)
			{
				calibrated = false;
				calibrateTimer += Time.deltaTime;
				if (connected)
				{
			    	calibrateButton.label = Mathf.Round(5.0f - calibrateTimer).ToString();
				}
				else
				{
					calibrateButton.label = "You need to connect first !";
					timerEnabled = false;	
					calibrateTimer = 0.0f;
				}
				if (calibrateTimer > 5.0f)
				{
					timerEnabled = false;
					calibrateTimer = 0.0f;
					calibrateButton.label = "Calibrate";
					if (connected)
					{
						SkyshopLightController sslc = SkyshopLightController.singleton;
						if (sslc != null)
							sslc.Flash();
						calibrateJoints();
					}
				}
			}

			if (udpSocket.Available > 1)
			{
				watchdogTimer = 0.0f;
				connected = true;
				status.val = "\nServer Status: <color=green>Connected</color>";
				udpSocket.ReceiveFrom(inBuffer, ref remote);
				// inBuffer byte array contains position, rotation, and tracking status data for all 20 Kinect360 joints

				int index = 0;
				// Get joint positions, inBuff[0-239]
				for (int i = 0; i < 20; i++) 
				{
					Decode3Float32(inBuffer, ref index, ref kinect.jointPositions[i]);
				}

				// Get joint rotations, inBuff[240-559]
				// Server is sending bone rotations but since they are unused, just bump the byte array index count past rotation data
				index+=320;//Disable this line and enable the next four to populate the raw joint rotation quaternion array.
				//for (int i = 0; i < 20; i++)
				//{
				//	Decode4Float32(inBuffer, ref index, ref kinect.jointRotations[i]);
				//}
				
				// Get joints tracking state, inBuff[560-639]
				for (int i = 0; i < 20; i++) 
				{
					DecodeFloat32(inBuffer, ref index, ref kinect.trackingState[i]);
				}

				if (calibrated && tracking.val && !selectNone)
				{
					for (int i = 0; i < 20; i++)
					{
						if (jointControls[i] != null && jointControls[i].currentPositionState != FreeControllerV3.PositionState.ParentLink)
						{
							if (kinect.track[i])
							{
								jointControls[i].currentRotationState = FreeControllerV3.RotationState.Off;	
								jointControls[i].currentPositionState = FreeControllerV3.PositionState.On;

								// Tracking State: 2 = Tracked, 1 = Inferred, 0 = NotTracked
								// Decrease joint spring if tracking is inferred, or not tracked
								jointControls[i].RBHoldPositionSpring = kinect.trackingState[i] == 2f ? 2000f
																	  : kinect.trackingState[i] == 1f ? 1000f
																	  : 500f;
								// Increase joint damping if tracking is inferred, or not tracked
								jointControls[i].RBHoldPositionDamper = kinect.trackingState[i] == 2f ? 35f
																	  : kinect.trackingState[i] == 1f ? 70f
																	  : 100f;

								jointControls[i].control.position = transform.TransformPoint(kinect.jointPositions[i] + kinect.offsets[i]);

							}
							else
							{	// Let non-tracked joints move freely relative to parent and gravity
								jointControls[i].currentRotationState = FreeControllerV3.RotationState.Off;
								jointControls[i].currentPositionState = FreeControllerV3.PositionState.Off;
							}
						}
					}
				}
			}
			else
			{	//Server watchdog timer
				watchdogTimer += Time.deltaTime;
				if (watchdogTimer > 5.0f && connected)
				{
					connected = false;
					calibrated = false;
					status.val = "\nServer Status: <color=red>Disconnected</color>";
					SuperController.LogMessage("Kinect2VAM server has disconnected.");
				}
			}
		}

		private void startCountdown()
		{
			timerEnabled = true;
		}

		private void calReset()
		{
			calibrated = false;
		}

		private void calibrateJoints()
		{
			for (int i = 0; i < 20; i++)
			{
				if (jointControls[i] != null && kinect.jointPositions[i] != null)
				{
					kinect.offsets[i] = transform.InverseTransformPoint(jointControls[i].control.position) - kinect.jointPositions[i];
				}
			}
			calibrated = true;
		}

        private void Decode3Float32(byte[] buffer, ref int index, ref Vector3 val)
        {
            val[0] = BitConverter.ToSingle(buffer, index);
            index += 4;
            val[1] = BitConverter.ToSingle(buffer, index);
            index += 4;
            val[2] = BitConverter.ToSingle(buffer, index);
            index += 4;
        }

        private void Decode4Float32(byte[] buffer, ref int index, ref Quaternion val)
        {
            val[0] = BitConverter.ToSingle(buffer, index);
            index += 4;
            val[1] = BitConverter.ToSingle(buffer, index);
            index += 4;
            val[2] = BitConverter.ToSingle(buffer, index);
            index += 4;
            val[3] = BitConverter.ToSingle(buffer, index);
            index += 4;	
        }

        private void DecodeFloat32(byte[] buffer, ref int index, ref float val)
        {
            val = BitConverter.ToSingle(buffer, index);
            index += 4;
        }

        void OnDisable()
        {
			calibrated = false;
			status.val = "\nServer Status: <color=red>Disconnected</color>";			
			udpSocket.SendTo(disconnectPacket, disconnectPacket.Length, SocketFlags.None, ipep);
			udpSocket.Close();
		}

		void OnEnable()
		{
			udpSocket = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
		}

		void OnDestroy()
		{
			udpSocket.SendTo(disconnectPacket, disconnectPacket.Length, SocketFlags.None, ipep);
			udpSocket.Close();
		}

		private void Connect()
		{
			udpSocket.SendTo(connectPacket, connectPacket.Length, SocketFlags.None, ipep);
			calibrateButton.label = "Calibrate";
		}
        
		private void Disconnect()
		{
			connected = false;
			calibrated = false;
			status.val = "\nServer Status: <color=red>Disconnected</color>";
			udpSocket.SendTo(disconnectPacket, disconnectPacket.Length, SocketFlags.None, ipep);
		}
			
		private void SelectAll()
		{
			for (int i = 0; i < 20; i++)
			{
				//Don't enable hands leaf node tracking. We only use hand.control (wrists).
				if (i != 7 && i != 11)
					kinect.track[i] = true;
			}		
			hip.val = true;
			chest.val = true;	
			pelvis.val = true;
			head.val = true;
			lShoulder.val = true;
			lElbow.val = true;
			lHand.val = true;
			rShoulder.val = true;
			rElbow.val = true;
			rHand.val = true;
			lThigh.val = true;
			lKnee.val = true;
			lFoot.val = true;
			lToe.val = true;
			rThigh.val = true;
			rKnee.val = true;
			rFoot.val = true;
			rToe.val = true;
			selectNone = false;
		}
		
		private void SelectNone()
		{
			selectNone = true;
			for (int i = 0; i < 20; i++)
			{
				kinect.track[i] = false;
			}
			hip.val = false;
			chest.val = false;	
			pelvis.val = false;
			head.val = false;
			lShoulder.val = false;
			lElbow.val = false;
			lHand.val = false;
			rShoulder.val = false;
			rElbow.val = false;
			rHand.val = false;
			lThigh.val = false;
			lKnee.val = false;
			lFoot.val = false;
			lToe.val = false;
			rThigh.val = false;
			rKnee.val = false;
			rFoot.val = false;
			rToe.val = false;
			SetJointStatesOn();
		}

		private void setTrackedJoints(bool val)
		{
			kinect.track[0] = hip.val;
			kinect.track[1] = chest.val;
			kinect.track[2] = pelvis.val;
			kinect.track[3] = head.val;
			kinect.track[4] = lShoulder.val;
			kinect.track[5] = lElbow.val;
			kinect.track[6] = lHand.val;
			kinect.track[8] = rShoulder.val;
			kinect.track[9] = rElbow.val;
			kinect.track[10] = rHand.val;
			kinect.track[12] = lThigh.val;
			kinect.track[13] = lKnee.val;
			kinect.track[14] = lFoot.val;
			kinect.track[15] = lToe.val;
			kinect.track[16] = rThigh.val;
			kinect.track[17] = rKnee.val;
			kinect.track[18] = rFoot.val;
			kinect.track[19] = rToe.val;

			bool allOff = true;
			for (int i = 0; i < 20; i++)
			{
				if (kinect.track[i] == true)
					allOff = false;
			}

			selectNone = allOff;
			if (selectNone)
				SetJointStatesOn();
		}

		private void SetJointStatesOn()
		{   //Set all but possessed joint position and rotation states to on when no single joint is selected for tracking.
			//This prevents the Person atom from falling to the floor (or into the abyss if there is no floor).
			for (int i = 0; i < 20; i++)
			{			
				if (jointControls[i] != null && jointControls[i].currentPositionState != FreeControllerV3.PositionState.ParentLink)	
				{					
					jointControls[i].currentRotationState = FreeControllerV3.RotationState.On;
					jointControls[i].currentPositionState = FreeControllerV3.PositionState.On;
				}
			}
		}

        private JSONStorableBool CreateBoolCheckbox(string name, bool defaultValue, bool rightSide)
        {
            JSONStorableBool storable = new JSONStorableBool(name, defaultValue, new JSONStorableBool.SetBoolCallback(setTrackedJoints));
            storable.storeType = JSONStorableParam.StoreType.Full;
            RegisterBool(storable);
            CreateToggle(storable, rightSide);
            return storable;
        }

        private JSONStorableFloat CreateFloatSlider(string name, float defaultValue, float minValue, float maxValue, bool rightSide)
        {
            JSONStorableFloat storable = new JSONStorableFloat(name, defaultValue, minValue, maxValue, true, true);
            storable.storeType = JSONStorableParam.StoreType.Full;
            CreateSlider(storable, rightSide);
            RegisterFloat(storable);
            return storable;
        }
    }

	public class Body
	{
		public Vector3[] jointPositions = new Vector3[20];
		public Quaternion[] jointRotations = new Quaternion[20];
		public float[] trackingState = new float[20];
		public Vector3[] offsets = new Vector3[20];
		public bool[] track = new bool[20];
	}
}